@@ -15,6 +15,7 @@
 #include "soup-message.h"
 #include "soup-message-private.h"
 #include "soup-socket.h"
+#include "soup-ssl.h"
 
 typedef enum {
 	SOUP_MESSAGE_IO_CLIENT,
@@ -76,6 +77,8 @@ io_cleanup (SoupMessage *msg)
 	if (!io)
 		return;
 
+	soup_message_io_stop (msg);
+
 	if (io->sock)
 		g_object_unref (io->sock);
 
@@ -89,6 +92,13 @@ io_cleanup (SoupMessage *msg)
 	msg->priv->io_data = NULL;
 }
 
+/**
+ * soup_message_io_stop:
+ * @msg: a #SoupMessage
+ *
+ * Immediately stops I/O on msg; if the connection would be left in an
+ * inconsistent state, it will be closed.
+ **/
 void
 soup_message_io_stop (SoupMessage *msg)
 {
@@ -123,7 +133,6 @@ static void
 soup_message_io_finished (SoupMessage *msg)
 {
 	g_object_ref (msg);
-	soup_message_io_stop (msg);
 	io_cleanup (msg);
 	if (SOUP_MESSAGE_IS_STARTING (msg))
 		soup_message_restarted (msg);
@@ -146,8 +155,18 @@ io_error (SoupSocket *sock, SoupMessage *msg)
 		return;
 	}
 
-	if (!SOUP_STATUS_IS_TRANSPORT_ERROR (msg->status_code))
-		soup_message_set_status (msg, SOUP_STATUS_IO_ERROR);
+	if (!SOUP_STATUS_IS_TRANSPORT_ERROR (msg->status_code)) {
+		GError *err = g_object_get_data (G_OBJECT (sock),
+						 "SoupSocket-last_error");
+
+		if (err && err->domain == SOUP_SSL_ERROR) {
+			soup_message_set_status_full (msg,
+						      SOUP_STATUS_SSL_FAILED,
+						      err->message);
+		} else
+			soup_message_set_status (msg, SOUP_STATUS_IO_ERROR);
+	}
+
 	soup_message_io_finished (msg);
 }
 
@@ -157,7 +176,8 @@ read_metadata (SoupMessage *msg, const char *boundary)
 	SoupMessageIOData *io = msg->priv->io_data;
 	SoupSocketIOStatus status;
 	char read_buf[RESPONSE_BLOCK_SIZE];
-	guint boundary_len = strlen (boundary), nread;
+	guint boundary_len = strlen (boundary);
+	gsize nread;
 	gboolean done;
 
 	do {
@@ -189,8 +209,9 @@ read_body_chunk (SoupMessage *msg)
 	SoupMessageIOData *io = msg->priv->io_data;
 	SoupSocketIOStatus status;
 	char read_buf[RESPONSE_BLOCK_SIZE];
-	guint nread, len = sizeof (read_buf);
+	guint len = sizeof (read_buf);
 	gboolean read_to_eof = (io->read_encoding == SOUP_TRANSFER_UNKNOWN);
+	gsize nread;
 
 	while (read_to_eof || io->read_length > 0) {
 		if (!read_to_eof)
@@ -240,7 +261,7 @@ write_data (SoupMessage *msg, const char *data, guint len)
 {
 	SoupMessageIOData *io = msg->priv->io_data;
 	SoupSocketIOStatus status;
-	guint nwrote;
+	gsize nwrote;
 
 	while (len > io->written) {
 		status = soup_socket_write (io->sock,
@@ -673,6 +694,18 @@ new_iostate (SoupMessage *msg, SoupSocket *sock, SoupMessageIOMode mode,
 	return io;
 }
 
+/**
+ * soup_message_io_client:
+ * @msg: a #SoupMessage
+ * @sock: socket to send @msg across
+ * @get_headers_cb: callback function to generate request headers
+ * @parse_headers_cb: callback function to parse response headers
+ * @user_data: data to pass to the callbacks
+ *
+ * Begins the process of sending @msg across @sock.
+ *
+ * Don't call this. Use soup_message_send_request().
+ **/
 void
 soup_message_io_client (SoupMessage *msg, SoupSocket *sock,
 			SoupMessageGetHeadersFn get_headers_cb,
@@ -691,6 +724,18 @@ soup_message_io_client (SoupMessage *msg, SoupSocket *sock,
 	io_write (sock, msg);
 }
 
+/**
+ * soup_message_io_server:
+ * @msg: an empty #SoupServerMessage
+ * @sock: socket to receive a request on
+ * @get_headers_cb: callback function to generate response headers
+ * @parse_headers_cb: callback function to parse request headers
+ * @user_data: data to pass to the callbacks
+ *
+ * Begins the process of receiving a request from @sock into @msg.
+ *
+ * Don't use this. Use soup_message_receive_request() instead.
+ **/
 void
 soup_message_io_server (SoupMessage *msg, SoupSocket *sock,
 			SoupMessageGetHeadersFn get_headers_cb,
@@ -709,6 +754,12 @@ soup_message_io_server (SoupMessage *msg, SoupSocket *sock,
 	io_read (sock, msg);
 }
 
+/**
+ * soup_message_io_pause:
+ * @msg: a #SoupMessage
+ *
+ * Pauses I/O on @msg.
+ **/
 void  
 soup_message_io_pause (SoupMessage *msg)
 {
@@ -726,6 +777,12 @@ soup_message_io_pause (SoupMessage *msg)
 	}
 }
 
+/**
+ * soup_message_io_unpause:
+ * @msg: a #SoupMessage
+ *
+ * Resumes I/O on @msg.
+ **/
 void  
 soup_message_io_unpause (SoupMessage *msg)
 {
